# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models
"""

from __future__ import unicode_literals

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declarative_base

from rattail.core import Object

# These are imported because most sibling modules import them *from* here (if
# that makes sense).  Probably need to think harder about the "best" place for
# these things to live.
from rattail.db.core import uuid_column, getset_factory
from rattail.db.types import GPCType


class ModelBase(Object):
    """
    Base class for all data models.
    """

    def __repr__(self):
        mapper = orm.object_mapper(self)
        keys = ['{0}={1}'.format(k.key, repr(getattr(self, k.key))) for k in mapper.primary_key]
        return "{0}({1})".format(self.__class__.__name__, ', '.join(keys))


Base = declarative_base(cls=ModelBase)


class Setting(Base):
    """
    Represents a setting stored within the database.
    """

    __tablename__ = 'setting'

    name = sa.Column(sa.String(length=255), primary_key=True)
    value = sa.Column(sa.Text())

    def __unicode__(self):
        return unicode(self.name or '')


class Change(Base):
    """
    Represents a changed (or deleted) record, which is pending synchronization
    to another database.
    """
    __tablename__ = 'change'

    uuid = uuid_column()

    class_name = sa.Column(sa.String(length=25), nullable=False)
    instance_uuid = sa.Column(sa.String(length=32), nullable=False)
    deleted = sa.Column(sa.Boolean(), nullable=False)

    def __repr__(self):
        return "Change(class_name={}, instance_uuid={}, deleted={})".format(
            repr(self.class_name), repr(self.instance_uuid), repr(self.deleted))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
SQLAlchemy-Continuum integration
"""

from __future__ import unicode_literals

import socket
import logging

import six

import sqlalchemy as sa
import sqlalchemy_continuum as continuum
from sqlalchemy_continuum.transaction import TransactionBase, create_triggers
from sqlalchemy_continuum.plugins import Plugin
from sqlalchemy_utils.functions import get_primary_keys

from rattail.util import OrderedDict

# TODO: Deprecate/remove this import.
from rattail.db.config import configure_versioning


log = logging.getLogger(__name__)


class RattailTransactionFactory(continuum.TransactionFactory):
    """
    Custom transaction factory for Continuum.  This is a little heavy-handed
    (it copies a fair chunk of code from upstream) but as of this writing was
    the only way to work around Continuum's assumption that the foreign key for
    ``Transaction.user`` should reference the ``user.id`` column (which we
    don't have).
    """

    def create_class(self, manager):
        """
        Create Transaction class.
        """
        class Transaction(
            manager.declarative_base,
            TransactionBase
        ):
            __tablename__ = 'transaction'
            __versioning_manager__ = manager

            id = sa.Column(
                sa.types.BigInteger,
                primary_key=True,
                autoincrement=True
            )

            if self.remote_addr:
                remote_addr = sa.Column(sa.String(50))

            if manager.user_cls:
                user_cls = manager.user_cls
                registry = manager.declarative_base._decl_class_registry

                if isinstance(user_cls, six.string_types):
                    try:
                        user_cls = registry[user_cls]
                    except KeyError:
                        raise continuum.ImproperlyConfigured(
                            'Could not build relationship between Transaction'
                            ' and %s. %s was not found in declarative class '
                            'registry. Either configure VersioningManager to '
                            'use different user class or disable this '
                            'relationship ' % (user_cls, user_cls)
                        )

                user_uuid = sa.Column(
                    sa.inspect(user_cls).primary_key[0].type,
                    sa.ForeignKey(
                        '%s.uuid' % user_cls.__tablename__
                    ),
                    index=True
                )

                user = sa.orm.relationship(user_cls)

            def __repr__(self):
                fields = ['id', 'issued_at', 'user']
                field_values = OrderedDict(
                    (field, getattr(self, field))
                    for field in fields
                    if hasattr(self, field)
                )
                return '<Transaction %s>' % ', '.join(
                    (
                        '%s=%r' % (field, value)
                        if not isinstance(value, six.integer_types)
                        # We want the following line to ensure that longs get
                        # shown without the ugly L suffix on python 2.x
                        # versions
                        else '%s=%d' % (field, value)
                        for field, value in field_values.items()
                    )
                )

        if manager.options['native_versioning']:
            create_triggers(Transaction)
        return Transaction


class RattailPlugin(Plugin):

    def transaction_args(self, uow, session):
        user = getattr(session, 'continuum_user', None)
        return {
            'user_uuid': user.uuid if user else None,
            'remote_addr': socket.gethostbyname(socket.gethostname()),
            }


versioning_manager = continuum.VersioningManager(
    transaction_cls=RattailTransactionFactory())


def disable_versioning(manager=None):
    """
    Disable Continuum versioning.
    """
    log.info("disabling Continuum versioning")
    if manager is None:
        manager = versioning_manager
    continuum.remove_versioning(manager=manager)


def count_versions(obj):
    """
    Return the number of versions given object has.

    .. note::
       This function was copied from the one at
       :func:`continuum:sqlalchemy_continuum.utils.count_versions()`, and
       changed so as not to try to use the ``repr()`` value of the table's
       primary key value, since for us that's usually an UUID as unicode string.

    :param obj: SQLAlchemy declarative model object
    """
    session = sa.orm.object_session(obj)
    if session is None:
        # If object is transient, we assume it has no version history.
        return 0
    manager = continuum.get_versioning_manager(obj)
    table_name = manager.option(obj, 'table_name') % obj.__table__.name

    def value(o, k):
        v = getattr(o, k)
        if isinstance(v, unicode):
            v = v.encode('utf_8')
        return v

    criteria = [
        '%s = %r' % (pk, value(obj, pk))
        for pk in get_primary_keys(obj)
    ]
    query = 'SELECT COUNT(1) FROM %s WHERE %s' % (
        table_name,
        ' AND '.join(criteria)
    )
    return session.execute(query).scalar()


def model_transaction_query(session, uuid, parent_class, child_classes=None):
    """
    Return a query capable of finding all Continuum ``Transaction`` instances
    which are associated with a model instance.
    """
    from rattail.db import model

    Transaction = continuum.transaction_class(parent_class)
    Version = continuum.version_class(parent_class)

    def normalize_child_classes():
        classes = []
        for cls in child_classes:
            if not isinstance(cls, tuple):
                cls = (cls, 'uuid')
            classes.append(cls)
        return classes

    query = session.query(Transaction)\
        .outerjoin(Version, sa.and_(
            Version.uuid == uuid,
            Version.transaction_id == Transaction.id))

    classes = [Version]
    if child_classes:
        for model_class, attr in normalize_child_classes():
            if isinstance(model_class, type) and issubclass(model_class, model.Base):
                cls = continuum.version_class(model_class)
                query = query.outerjoin(cls, sa.and_(
                        cls.transaction_id == Transaction.id,
                        getattr(cls, attr) == uuid))
                classes.append(cls)

    return query.filter(sa.or_(
            *[cls.uuid != None for cls in classes]))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Batch Makers
"""

from __future__ import unicode_literals

from rattail.db.batches.types import get_batch_type
from rattail.db.batches.data import BatchDataProvider
# from rattail.db.model import Batch, LabelProfile
from ..model import Batch
from rattail.sil import consume_batch_id


class BatchMaker(object):

    progress_message = "Making batch(es)"

    def __init__(self, session, source=None):
        self.session = session
        self.source = source
        self.batches = {}

    def get_batch(self, name):
        if name not in self.batches:
            self.batches[name] = self.make_batch(name)
        return self.batches[name]

    def make_batch(self, name):
        if hasattr(self, 'make_batch_%s' % name):
            batch = getattr(self, 'make_batch_%s' % name)()

        else:
            batch_type = get_batch_type(name)
            batch = Batch()
            batch_type.initialize(batch)
            if self.source:
                batch.source = self.source
            batch.id = consume_batch_id()

        self.session.add(batch)
        self.session.flush()
        batch.create_table()
        return batch

    def make_batches_begin(self, data):
        pass

    def make_batches(self, data, progress=None):
        if not isinstance(data, BatchDataProvider):
            raise TypeError("Sorry, you must pass a BatchDataProvider instance")

        result = self.make_batches_begin(data)
        if result is not None and not result:
            return False

        prog = None
        if progress and len(data):
            prog = progress(self.progress_message, len(data))

        cancel = False
        for i, data_row in enumerate(data, 1):
            self.process_data_row(data_row)
            if prog and not prog.update(i):
                cancel = True
                break
            self.session.flush()

        if prog:
            prog.destroy()

        if not cancel:
            result = self.make_batches_end()
            if result is not None:
                cancel = not result

        return not cancel

    def make_batches_end(self):
        pass

    def process_data_row(self, data_row):
        raise NotImplementedError


# class LabelsBatchMaker(BatchMaker):

#     default_profile = None
#     default_quantity = 1

#     def make_batches_begin(self, data):
#         if not self.default_profile:
#             q = self.session.query(LabelProfile)
#             q = q.order_by(LabelProfile.ordinal)
#             self.default_profile = q.first()
#             assert self.default_profile

#     def process_data_row(self, data_row):
#         batch = self.get_batch('labels')
#         row = batch.rowclass()
#         row.F01 = data_row.F01
#         row.F155 = data_row.F155
#         row.F02 = data_row.F02
#         row.F22 = data_row.F22
#         row.F95 = self.default_profile.code
#         row.F94 = self.default_quantity
#         batch.add_row(row)        



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor catalog parser for United Natural Foods (UNFI)
"""

from __future__ import unicode_literals

import re
import datetime
from decimal import Decimal

import xlrd

from rattail.vendors.catalogs import CatalogParser
from rattail.db.batch.vendorcatalog import VendorCatalogRow
from rattail.gpc import GPC


class UNFICatalogParser(CatalogParser):
    """
    Vendor catalog parser for UNFI, version 1.
    """
    key = 'rattail.contrib.unfi'
    display = "United Natural Foods (UNFI), v1"
    vendor_key = 'unfi'

    effective_date_pattern = re.compile(r'effective from (\d{2}/\d{2}/\d{4})-')
    code_pattern = re.compile(r'^\d{6}-\d$')
    upc_pattern = re.compile(r'^\d-\d{5}-\d{5}-\d$')

    def open_sheet(self, path):
        book = xlrd.open_workbook(path)
        return book.sheet_by_index(0)

    def parse_effective_date(self, path):
        sheet = self.open_sheet(path)
        text = sheet.cell_value(0, 0)
        match = self.effective_date_pattern.search(text)
        if match:
            return datetime.datetime.strptime(match.group(1), '%m/%d/%Y').date()

    def parse_rows(self, path):
        sheet = self.open_sheet(path)
        for r in range(sheet.nrows):

            # Use the code to determine if the row is valid.
            code = sheet.cell_value(r, 0)
            if not self.code_pattern.match(code):
                continue

            # Warn if UPC is not valid.
            upc = sheet.cell_value(r, 1)
            if self.upc_pattern.match(upc):
                upc = GPC(upc.replace('-', ''))
            else:
                log.warning("invalid upc at row {0}: {1}".format(r + 1, upc))
                upc = None

            row = VendorCatalogRow()
            row.upc = upc
            row.vendor_code = code.replace('-', '')
            row.brand_name = sheet.cell_value(r, 2)
            row.description = sheet.cell_value(r, 3)
            row.size = sheet.cell_value(r, 5)
            row.case_size = int(sheet.cell_value(r, 4))
            row.case_cost = Decimal(unicode(sheet.cell_value(r, 8)))
            row.unit_cost = Decimal(unicode(sheet.cell_value(r, 9)))
            yield row


class UNFICatalogParser2(UNFICatalogParser):
    """
    Vendor catalog parser for UNFI, version 2.
    """
    key = 'rattail.contrib.unfi.2'
    display = "United Natural Foods (UNFI), v2"

    def parse_rows(self, path):
        sheet = self.open_sheet(path)
        for r in range(sheet.nrows):

            # Use the code to determine if the row is valid.
            code = sheet.cell_value(r, 0)
            if not self.code_pattern.match(code):
                continue

            # Warn if UPC is not valid.
            upc = sheet.cell_value(r, 1)
            if self.upc_pattern.match(upc):
                upc = GPC(upc.replace('-', ''))
            else:
                log.warning("invalid upc at row {0}: {1}".format(r + 1, upc))
                upc = None

            row = VendorCatalogRow()
            row.upc = upc
            row.vendor_code = code.replace('-', '')
            row.brand_name = sheet.cell_value(r, 2)
            row.description = sheet.cell_value(r, 4)
            row.size = sheet.cell_value(r, 7)
            row.case_size = int(sheet.cell_value(r, 6))
            row.case_cost = Decimal(unicode(sheet.cell_value(r, 10)))
            row.unit_cost = Decimal(unicode(sheet.cell_value(r, 11)))
            yield row



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail Data Normalization
"""

from __future__ import unicode_literals, absolute_import


class Normalizer(object):
    """
    Base class for data normalizers.
    """

    def normalize(self, instance):
        raise NotImplementedError


class UserNormalizer(Normalizer):
    """
    Normalizer for user data.
    """
    # Must set this to the administrator Role instance.
    admin = None

    def normalize(self, user):
        return {
            'uuid': user.uuid,
            'username': user.username,
            'password': user.password,
            'salt': user.salt,
            'person_uuid': user.person_uuid,
            'active': user.active,
            'admin': self.admin in user.roles,
        }


class DepartmentNormalizer(Normalizer):
    """
    Normalizer for department data.
    """

    def normalize(self, department):
        return {
            'uuid': department.uuid,
            'number': department.number,
            'name': department.name,
        }


class EmployeeNormalizer(Normalizer):
    """
    Normalizer for employee data.
    """

    def normalize(self, employee):
        data = {
            'uuid': employee.uuid,
            'id': employee.id,
            'person_uuid': employee.person_uuid,
            'status': employee.status,
            'first_name': employee.first_name,
            'last_name': employee.last_name,
            'display_name': employee.display_name,
        }

        customer = employee.customers[0] if employee.customers else None
        data['customer_id'] = customer.id if customer else None

        data['phone_number'] = None
        for phone in employee.phones:
            if phone.type == 'Home':
                data['phone_number'] = phone.number
                break

        data['phone_number_2'] = None
        first = False
        for phone in employee.phones:
            if phone.type == 'Home':
                if first:
                    data['phone_number_2'] = phone.number
                    break
                first = True

        email = employee.email
        data['email_address'] = email.address if email else None

        return data


class EmployeeStoreNormalizer(Normalizer):
    """
    Normalizer for employee_x_store data.
    """

    def normalize(self, emp_store):
        return {
            'uuid': emp_store.uuid,
            'employee_uuid': emp_store.employee_uuid,
            'store_uuid': emp_store.store_uuid,
        }


class EmployeeDepartmentNormalizer(Normalizer):
    """
    Normalizer for employee_x_department data.
    """

    def normalize(self, emp_dept):
        return {
            'uuid': emp_dept.uuid,
            'employee_uuid': emp_dept.employee_uuid,
            'department_uuid': emp_dept.department_uuid,
        }


class MessageNormalizer(Normalizer):
    """
    Normalizer for message data.
    """

    def normalize(self, message):
        return {
            'uuid': message.uuid,
            'sender_uuid': message.sender_uuid,
            'subject': message.subject,
            'body': message.body,
            'sent': message.sent,
        }


class MessageRecipientNormalizer(Normalizer):
    """
    Normalizer for message recipient data.
    """

    def normalize(self, recip):
        return {
            'uuid': recip.uuid,
            'message_uuid': recip.message_uuid,
            'recipient_uuid': recip.recipient_uuid,
            'status': recip.status,
        }



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Configuration
"""

from __future__ import unicode_literals
from __future__ import absolute_import

from rattail.config import ConfigProfile, parse_list
from rattail.util import load_object
from rattail.exceptions import ConfigurationError


class DataSyncProfile(ConfigProfile):
    """
    Simple class to hold configuration for a DataSync "profile".  Each profile
    determines which database(s) will be watched for new changes, and which
    consumer(s) will then be instructed to process the changes.

    .. todo::
       This clearly needs more documentation.
    """
    section = 'rattail.datasync'

    def __init__(self, config, key):
        self.config = config
        self.key = key

        self.watcher_spec = self._config_string('watcher')
        self.watcher = load_object(self.watcher_spec)(config, key)
        self.watcher.delay = self._config_int('watcher.delay', default=self.watcher.delay)

        consumers = self._config_list('consumers')
        if consumers == ['self']:
            self.watcher.consumes_self = True
        else:
            self.watcher.consumes_self = False
            self.consumers = self.normalize_consumers(consumers)
            self.consumer_delay = self._config_int('consumer_delay', default=1)
            self.watcher.consumer_stub_keys = [c.key for c in self.isolated_consumers]
            if self.common_consumers:
                self.watcher.consumer_stub_keys.append(None)

    @property
    def isolated_consumers(self):
        return [c for c in self.consumers if c.isolated]

    @property
    def common_consumers(self):
        return [c for c in self.consumers if not c.isolated]

    def normalize_consumers(self, raw_consumers):
        # TODO: This will do for now, but isn't as awesome as it probably could be.
        default_isolated = self._config_boolean('consumers.isolated', default=True)
        consumers = []
        for key in raw_consumers:
            consumer_spec = self._config_string('consumer.{0}'.format(key))
            dbkey = self._config_string('consumer.{0}.db'.format(key), default=key)
            consumer = load_object(consumer_spec)(self.config, key, dbkey=dbkey)
            consumer.spec = consumer_spec
            consumer.isolated = self._config_boolean('consumer.{0}.isolated'.format(key),
                                                     default=default_isolated)
            consumers.append(consumer)
        return consumers


def get_profile_keys(config):
    """
    Returns a list of profile keys used in the DataSync configuration.
    """
    keys = config.get('rattail.datasync', 'watch')
    if keys:
        return parse_list(keys)


def load_profiles(config):
    """
    Load all active DataSync profiles defined within configuration.
    """
    # Make sure we have a top-level directive.
    keys = get_profile_keys(config)
    if not keys:
        raise ConfigurationError(
            "The DataSync configuration does not specify any profiles "
            "to be watched.  Please defined the 'watch' option within "
            "the [rattail.datasync] section of your config file.")

    watched = {}
    for key in keys:
        profile = DataSyncProfile(config, key)
        watched[key] = profile
    return watched



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Users & Permissions
"""

from __future__ import unicode_literals

import sqlalchemy as sa
from sqlalchemy.orm import relationship
from sqlalchemy.ext.associationproxy import association_proxy

from .core import Base, uuid_column, getset_factory
from .people import Person
 

class Role(Base):
    """
    Represents a role within the system; used to manage permissions.
    """
    __tablename__ = 'role'
    __table_args__ = (
        sa.UniqueConstraint('name', name='role_uq_name'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    name = sa.Column(sa.String(length=25), nullable=False)

    def __unicode__(self):
        return unicode(self.name or '')


class Permission(Base):
    """
    Represents permission a role has to do a particular thing.
    """
    __tablename__ = 'permission'
    __table_args__ = (
        sa.ForeignKeyConstraint(['role_uuid'], ['role.uuid'], name='permission_fk_role'),
        )

    role_uuid = sa.Column(sa.String(length=32), primary_key=True)
    permission = sa.Column(sa.String(length=50), primary_key=True)

    def __unicode__(self):
        return unicode(self.permission or '')


Role._permissions = relationship(
    Permission, backref='role',
    cascade='save-update, merge, delete, delete-orphan')

Role.permissions = association_proxy(
    '_permissions', 'permission',
    creator=lambda p: Permission(permission=p),
    getset_factory=getset_factory)


class User(Base):
    """
    Represents a user of the system.

    This may or may not correspond to a real person, i.e. some users may exist
    solely for automated tasks.
    """
    __tablename__ = 'user'
    __table_args__ = (
        sa.ForeignKeyConstraint(['person_uuid'], ['person.uuid'], name='user_fk_person'),
        sa.UniqueConstraint('username', name='user_uq_username'),
        )
    __versioned__ = {'exclude': ['password', 'salt']}

    uuid = uuid_column()
    username = sa.Column(sa.String(length=25), nullable=False)
    password = sa.Column(sa.String(length=60))
    salt = sa.Column(sa.String(length=29))
    person_uuid = sa.Column(sa.String(length=32))

    active = sa.Column(sa.Boolean(), nullable=False, default=True)
    """
    Whether the user is active, e.g. allowed to log in via the UI.
    """

    def __unicode__(self):
        s = unicode(self.username)
        if self.person and self.person.display_name:
            s = "{0} ({1})".format(s, self.person.display_name)
        return s

    @property
    def display_name(self):
        """
        Display name for the user.
        
        Returns :attr:`Person.display_name` if available; otherwise returns
        :attr:`username`.
        """
        if self.person and self.person.display_name:
            return self.person.display_name
        return self.username

    @property
    def employee(self):
        """
        Reference to the :class:`Employee` associated with the user, if any.
        """
        if self.person:
            return self.person.employee

    def get_email_address(self):
        """
        Returns the primary email address for the user (as unicode string), or
        ``None``.  Note that currently there is no direct association between a
        User and an EmailAddress, so the Person and Customer relationships are
        navigated in an attempt to locate an address.
        """
        if self.person:
            if self.person.email:
                return self.person.email.address
            for customer in self.person.customers:
                if customer.email:
                    return customer.email.address

    @property
    def email_address(self):
        """
        Convenience attribute which invokes :meth:`get_email_address()`.

        .. note::
           The implementation of this may change some day, e.g. if the User is
           given an association to EmailAddress in the data model.
        """
        return self.get_email_address()


User.person = relationship(
    Person,
    back_populates='user',
    uselist=False)

Person.user = relationship(
    User,
    back_populates='person',
    uselist=False)


class UserRole(Base):
    """
    Represents the association between a :class:`User` and a :class:`Role`.
    """
    __tablename__ = 'user_x_role'
    __table_args__ = (
        sa.ForeignKeyConstraint(['user_uuid'], ['user.uuid'], name='user_x_role_fk_user'),
        sa.ForeignKeyConstraint(['role_uuid'], ['role.uuid'], name='user_x_role_fk_role'),
        )

    uuid = uuid_column()
    user_uuid = sa.Column(sa.String(length=32), nullable=False)
    role_uuid = sa.Column(sa.String(length=32), nullable=False)


Role._users = relationship(
    UserRole, backref='role',
    cascade='all, delete-orphan')

Role.users = association_proxy(
    '_users', 'user',
    creator=lambda u: UserRole(user=u),
    getset_factory=getset_factory)

User._roles = relationship(
    UserRole, backref='user',
    cascade='all, delete-orphan')

User.roles = association_proxy(
    '_roles', 'role',
    creator=lambda r: UserRole(role=r),
    getset_factory=getset_factory)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2014 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Global Product Code
"""

from __future__ import unicode_literals

from rattail import barcodes


class GPC(object):
    """
    Class to abstract the details of Global Product Code data.  Examples of
    this would be UPC or EAN barcodes.

    The initial motivation for this class was to provide better SIL support.
    To that end, the instances are assumed to always be comprised of only
    numeric digits, and must include a check digit.  If you do not know the
    check digit, provide a ``calc_check_digit`` value to the constructor.
    """

    def __init__(self, value, calc_check_digit=False):
        """
        Constructor.  ``value`` must be either an integer or a long value, or a
        string containing only digits.

        If ``calc_check_digit`` is ``False``, then ``value`` is assumed to
        include the check digit.  If the value does not include a check digit
        and needs one to be calculated, then ``calc_check_digit`` should be a
        keyword signifying the algorithm to be used.

        Currently the only check digit algorithm keyword supported is
        ``'upc'``.  As that is likely to always be the default, a
        ``calc_check_digit`` value of ``True`` will be perceived as equivalent
        to ``'upc'``.
        """

        value = str(value)
        if calc_check_digit is True or calc_check_digit == 'upc':
            value += str(barcodes.upc_check_digit(value))
        self.value = int(value)

    def __eq__(self, other):
        try:
            return int(self) == int(other)
        except (TypeError, ValueError):
            return False

    def __ne__(self, other):
        try:
            return int(self) != int(other)
        except (TypeError, ValueError):
            return True

    def __cmp__(self, other):
        if int(self) < int(other):
            return -1
        if int(self) > int(other):
            return 1
        if int(self) == int(other):
            return 0
        assert False

    def __hash__(self):
        return hash(self.value)

    def __int__(self):
        return int(self.value)

    def __long__(self):
        return long(self.value)

    def __repr__(self):
        return "GPC('%014d')" % self.value

    def __str__(self):
        return str(unicode(self))

    def __unicode__(self):
        return u'%014d' % self.value

    def pretty(self):
        """
        Returns the UPC as a somewhat more human-readable string.  Basically
        that just means the check digit is distinguished by a hyphen.
        """
        upc = unicode(self)
        return "{0}-{1}".format(upc[:-1], upc[-1])



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
DataSync Utilities
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm


def make_dependency_sorter(model):
    """
    Return a function suitable for use when sorting data model class names,
    according to the underlying foreign key dependencies between the tables.
    """

    def dependency_sorter(x, y):
        map_x = orm.class_mapper(getattr(model, x))
        map_y = orm.class_mapper(getattr(model, y))

        dep_x = []
        table_y = map_y.tables[0].name
        for column in map_x.columns:
            for key in column.foreign_keys:
                if key.column.table.name == table_y:
                    return 1
                dep_x.append(key)

        dep_y = []
        table_x = map_x.tables[0].name
        for column in map_y.columns:
            for key in column.foreign_keys:
                if key.column.table.name == table_x:
                    return -1
                dep_y.append(key)

        if dep_x and not dep_y:
            return 1
        if dep_y and not dep_x:
            return -1
        return 0

    return dependency_sorter



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
CSV Import Data Providers
"""

from __future__ import unicode_literals

import datetime
from decimal import Decimal

from .core import DataProvider
from rattail.db import model
from rattail.gpc import GPC
from rattail.db.importing import models
from rattail.util import load_object
from rattail.csvutil import UnicodeDictReader
from rattail.db.util import maxlen
from rattail.time import localtime, make_utc


def make_provider(config, session, spec, **kwargs):
    """
    Create a provider instance according to the given spec.  For now..see the
    source code for more details.
    """
    provider = None
    if '.' not in spec and ':' not in spec:
        from rattail.db.importing.providers import csv
        if hasattr(csv, spec):
            provider = getattr(csv, spec)
        elif hasattr(csv, '{0}Provider'.format(spec)):
            provider = getattr(csv, '{0}Provider'.format(spec))
    else:
        provider = load_object(spec)
    if provider:
        return provider(config, session, **kwargs)


class CsvProvider(DataProvider):
    """
    Base class for CSV data providers.
    """
    time_format = '%Y-%m-%d %H:%M:%S'

    def get_source_data(self, progress=None):
        with open(self.data_path, 'rb') as f:
            reader = UnicodeDictReader(f)
            return list(reader)

    def make_utc(self, time):
        if time is None:
            return None
        return make_utc(localtime(self.config, time))

    def make_time(self, value):
        if not value:
            return None
        time = datetime.datetime.strptime(value, self.time_format)
        return self.make_utc(time)


class ProductProvider(CsvProvider):
    """
    CSV product data provider.
    """
    importer_class = models.ProductImporter
    supported_fields = [
        'uuid',
        'upc',
        'description',
        'size',
        'department_uuid',
        'subdepartment_uuid',
        'category_uuid',
        'brand_uuid',
        'regular_price',
        'sale_price',
        'sale_starts',
        'sale_ends',
        ]
    maxlen_description = maxlen(model.Product.description)
    maxlen_size = maxlen(model.Product.size)

    def normalize(self, data):

        if 'upc' in data:
            upc = data['upc']
            data['upc'] = GPC(upc) if upc else None

        # Warn about truncation until Rattail schema is addressed.
        if 'description' in data:
            description = data['description'] or ''
            if self.maxlen_description and len(description) > self.maxlen_description:
                log.warning("product description is more than {} chars and will be truncated: {}".format(
                        self.maxlen_description, repr(description)))
                description = description[:self.maxlen_description]
            data['description'] = description

        # Warn about truncation until Rattail schema is addressed.
        if 'size' in data:
            size = data['size'] or ''
            if self.maxlen_size and len(size) > self.maxlen_size:
                log.warning("product size is more than {} chars and will be truncated: {}".format(
                        self.maxlen_size, repr(size)))
                size = size[:self.maxlen_size]
            data['size'] = size

        if 'department_uuid' in data:
            data['department_uuid'] = data['department_uuid'] or None

        if 'subdepartment_uuid' in data:
            data['subdepartment_uuid'] = data['subdepartment_uuid'] or None

        if 'category_uuid' in data:
            data['category_uuid'] = data['category_uuid'] or None

        if 'brand_uuid' in data:
            data['brand_uuid'] = data['brand_uuid'] or None

        if 'regular_price' in data:
            price = data['regular_price']
            data['regular_price'] = Decimal(price) if price else None

        # Determine if sale price is currently active; if it is not then we
        # will declare None for all sale fields.
        if 'sale_starts' in data:
            data['sale_starts'] = self.make_time(data['sale_starts'])
        if 'sale_ends' in data:
            data['sale_ends'] = self.make_time(data['sale_ends'])
        if 'sale_price' in data:
            price = data['sale_price']
            data['sale_price'] = Decimal(price) if price else None
            if data['sale_price']:
                sale_starts = data.get('sale_starts')
                sale_ends = data.get('sale_ends')
                active = False
                if sale_starts and sale_ends:
                    if sale_starts <= self.now <= sale_ends:
                        active = True
                elif sale_starts:
                    if sale_starts <= self.now:
                        active = True
                elif sale_ends:
                    if self.now <= sale_ends:
                        active = True
                else:
                    active = True
                if not active:
                    data['sale_price'] = None
                    data['sale_starts'] = None
                    data['sale_ends'] = None
            else:
                data['sale_starts'] = None
                data['sale_ends'] = None

        return data
